/*
 * Copyright (C) 2011-2012 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//
// Module Name: alu4
//   ver 001
// 
// ALU based 74181 & 74182
//
// (1) normalized C_IN , C_OUT
// (2) enhanced Function :
//
//     S[7:4]   S[3:0]
//
//     1001      FS[3:0]  -- normal 181
//     0110      FS[3:0]  -- invert A
//     FS[3:0]   1001     -- swap A,B
//     FS[3:0]   0110     -- invert B and swap A,B
//
//
//   FS[3:0]      M = 1      C_IN = 0    C_IN = 1
//                          (no carry)  (with carry)
//   0 0 0 0       ~A            A          A + 1
//   0 0 0 1    ~(A | B)  
//   0 0 1 0     ~A & B  
//   0 0 1 1        0           -1            0
//   0 1 0 0    ~(A & B)   
//   0 1 0 1       ~B    
//   0 1 1 0      A ^ B        A - B - 1    A - B   
//   0 1 1 1      A & ~B       
//   1 0 0 0     ~A | B
//   1 0 0 1    ~(A ^ B)       A + B        A + B + 1
//   1 0 1 0        B           
//   1 0 1 1      A & B
//   1 1 0 0        1          A + A        A + A + 1
//   1 1 0 1      A | ~B
//   1 1 1 0      A | B
//   1 1 1 1        A          A - 1          A
//
module clg182 (
		 input   C_IN       // Carry Input
		 , input   P0
		 , input   G0
		 , input   P1
		 , input   G1
		 , input   P2
		 , input   G2
		 , input   P3
		 , input   G3
		 , output   C_OUT1       // Carry Output
		 , output   C_OUT2       // Carry Output
		 , output   C_OUT3       // Carry Output
		 , output   P
		 , output   G
		);

	assign C_OUT1 = G0 & ( P0 | C_IN );
	assign C_OUT2 = G1 & ( P1 | G0 & ( P0 | C_IN ));
	assign C_OUT3 = G2 & ( P2 | G1 & ( P1 | G0 & (P0 | C_IN )));

	assign G = G3 & ( P3 | G2 & ( P2 | G1 & (P1 & G0 )));
	assign P = P0 | P1 | P2 | P3;
endmodule

module alu32 (
		 input   C_IN       // Carry Input
		, input  [31:0] A     // DATA IN
		, input  [31:0] B     // DATA IN
		, output [31:0] F   // DATA OUT
		, output  OVR      // Overflow
		, output  EQ       // F != 0
		, output  C_OUT      // carry-look-ahead

		, output  P          // carry-propagate (INV)
		, output  G          // carry-generate (INV)

		, input   M          // Mode Control
		, input   [7:0] S   // Function Select
		);

	clg182 clg (.C_IN(C_IN) 
		     , .P0(P0), .G0(G0), .P1(P1), .G1(G1)
		     , .P2(P2), .G2(G2), .P3(P3), .G3(G3)
		     , .C_OUT1(C_OUT1), .C_OUT2(C_OUT2), .C_OUT3(C_OUT3)
		     , .P(P), .G(G) );

	alu8 alu_0 (.C_IN(C_IN), .A(A[7:0]), .B(B[7:0]), .F(F[7:0])
		     , .M(M), .S(S[7:0]), .C_OUT(C_OUT0)
		     , .EQ(EQ_0) , .P(P0), .G(G0) );
	alu8 alu_1 (.C_IN(C_OUT1), .A(A[15:8]), .B(B[15:8]), .F(F[15:8])
		     , .M(M), .S(S[7:0])
		     , .EQ(EQ_1) , .P(P1), .G(G1) );
	alu8 alu_2 (.C_IN(C_OUT2), .A(A[23:16]), .B(B[23:16]), .F(F[23:16])
		     , .M(M), .S(S[7:0])
		     , .EQ(EQ_2) , .P(P2), .G(G2) );
	alu8 alu_3 (.C_IN(C_OUT3), .A(A[31:24]), .B(B[31:24]), .F(F[31:24])
		     , .M(M), .S(S[7:0]), .C_OUT(C_OUT) , .OVR(OVR)
		     , .EQ(EQ_3) , .P(P3), .G(G3) );

	assign EQ = EQ_0 & EQ_1 & EQ_2 & EQ_3;
endmodule
