/*
 * Copyright (C) 2011-2012 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//
// Module Name: alu4
//   ver 001
// 
// ALU based 74181
//
// (1) normalized C_IN , C_OUT
// (2) enhanced Function :
//
//     S[7:4]   S[3:0]
//
//     1001      FS[3:0]  -- normal 181
//     0110      FS[3:0]  -- invert A
//     FS[3:0]   1001     -- swap A,B
//     FS[3:0]   0110     -- invert B and swap A,B
//
//
//   FS[3:0]      M = 1      C_IN = 0    C_IN = 1
//                          (no carry)  (with carry)
//   0 0 0 0       ~A            A          A + 1
//   0 0 0 1    ~(A | B)  
//   0 0 1 0     ~A & B  
//   0 0 1 1        0           -1            0
//   0 1 0 0    ~(A & B)   
//   0 1 0 1       ~B    
//   0 1 1 0      A ^ B        A - B - 1    A - B   
//   0 1 1 1      A & ~B       
//   1 0 0 0     ~A | B
//   1 0 0 1    ~(A ^ B)       A + B        A + B + 1
//   1 0 1 0        B           
//   1 0 1 1      A & B
//   1 1 0 0        1          A + A        A + A + 1
//   1 1 0 1      A | ~B
//   1 1 1 0      A | B
//   1 1 1 1        A          A - 1          A
//

module alu4 (
		 input   C_IN      // Carry Input
		, input  [3:0] A   // DATA IN
		, input  [3:0] B   // DATA IN
		, output [3:0] F   // DATA OUT
		, output  OVR      // Overflow
		, output  EQ       // F != 0
		, output  C_OUT    // carry-look-ahead

		, input   M        // Mode Control
		, input   [7:0] S  // Function Select

		, output  P        // carry-propagate (INV)
		, output  G        // carry-generate (INV)
		);

    wire [3:0] D;
    wire [3:0] C;
    assign D[0] = ~( A[0] & S[4] | ~A[0] & S[5] | B[0] & S[0] | ~B[0] & S[1] );
    assign D[1] = ~( A[1] & S[4] | ~A[1] & S[5] | B[1] & S[0] | ~B[1] & S[1] );
    assign D[2] = ~( A[2] & S[4] | ~A[2] & S[5] | B[2] & S[0] | ~B[2] & S[1] );
    assign D[3] = ~( A[3] & S[4] | ~A[3] & S[5] | B[3] & S[0] | ~B[3] & S[1] );

    assign C[0] = ~((~A[0] & S[6] | A[0] & S[7])&(~B[0] & S[2] | B[0] & S[3]));
    assign C[1] = ~((~A[1] & S[6] | A[1] & S[7])&(~B[1] & S[2] | B[1] & S[3]));
    assign C[2] = ~((~A[2] & S[6] | A[2] & S[7])&(~B[2] & S[2] | B[2] & S[3]));
    assign C[3] = ~((~A[3] & S[6] | A[3] & S[7])&(~B[3] & S[2] | B[3] & S[3]));

    wire C_OUT1 = ~( D[0] | ~C_IN & C[0]);
    wire C_OUT2 = ~( D[1] | C[1] & D[0] | ~C_IN & C[0] & C[1]);
    wire C_OUT3 = ~( D[2] | C[2] & ( D[1] | C[1] & D[0])
		     | ~C_IN & C[0] & C[1] & C[2]);

    assign P = ~( C[3] & C[2] & C[1] & C[0] );
    assign G = ~(D[3] | C[3] & ( D[2] | C[2] & ( D[1] | C[1] & D[0])));
    assign C_OUT = ~(~G | ~C_IN & C[0] & C[1] & C[2] & C[3]);

    assign F[0] = ~( ~M & ~C_IN ) ^ D[0] ^ C[0];
    assign F[1] = ~( ~M & ~C_OUT1 ) ^ D[1] ^ C[1];
    assign F[2] = ~( ~M & ~C_OUT2 ) ^ D[2] ^ C[2];
    assign F[3] = ~( ~M & ~C_OUT3 ) ^ D[3] ^ C[3];

    assign OVR = ~M & (C_OUT3 ^ C_OUT);
    assign EQ = F[0] & F[1] & F[2] & F[3];
endmodule
