/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
SPI timing:

SCK
          ___     ___     ___     ___     ___     ___     ___     ___   
   ______|   |___|   |___|   |___|   |___|   |___|   |___|   |___|   |_

r_count[2:0]

 |   0   |   1   |  2    |   3   |   4   |   5   |   6   |  7   | 

MOSI
     < bit7 >< bit6 >< bit5 >< bit4 >< bit3 >< bit2 >< bit1 >< bit0 >

r_valid                                                           ______
  _______________________________________________________________|      |____

r_req  _                                                      _
  ____| |____________________________________________________| |________
timing with CLK :

CLK
      ___     ___     ___     ___     ___     ___     ___     ___     ___
   __|   |___|   |___|   |___|   |___|   |___|   |___|   |___|   |___|   |___|

         :                               :
SCK      :                               :
   _____                                 _______________________________   
        |_______________________________|                               |_______   

         :                               :
r_valid  :                               :
                                          _________________________________
_________________________________________|                                 
         :                               :
r_req    :                               :
          _______
_________|       |_________________________________________________________

                 @ SPI_DATA_I input 

  Version 1.0
  Version 1.1
	add SPI_DATA_ACK
*/

module jtag_spi (
		input CLK
		, input MOSI
		, input SCK
		, input CS
		, output MISO

		, input [7:0] SPI_DATA_I
		, output SPI_DATA_REQ
		, input SPI_DATA_ACK
		, output [7:0] SPI_DATA
		, output SPI_DATA_VALID
		, output [1:0] SPI_DATA_COUNT
		);

    reg r_sck = 1'b0;
    wire r_spi_in = MOSI;

    reg [4:0]  r_count = 0;
    reg [1:0]  r_count_out;
    reg [7:0]  r_spi;
    reg [7:0]  r_spi_out;
    reg [7:0]  r_data; // recieved data
    reg r_data_valid = 0;
    reg r_data_req = 0;
    
    assign SPI_DATA       = r_spi[7:0];
    assign SPI_DATA_VALID = r_data_valid;
    assign SPI_DATA_REQ   = r_data_req;
    assign SPI_DATA_COUNT = r_count_out[1:0];

    assign MISO = r_spi_out[7];

    always @(negedge CLK)
    begin
	r_sck       <= SCK;

	if (CS)
	    begin
 		r_count      <=  0;
 		r_data_valid <= 1'b0;
 		r_data_req   <= 1'b0;
		r_spi_out    <= 0;
	    end
	else if (~r_sck & SCK) // posedge SCK : INPUT
	    begin
		r_count <= r_count + 1;
		r_spi <= {r_spi[6:0], r_spi_in };

		if (r_count[2:0] == 7)  r_count_out[1:0] <= r_count[4:3];
		if (r_count[2:0] == 7)  r_data_valid <= 1'b1;
		else                    r_data_valid <= 1'b0;
	    end
	else if (r_sck & ~SCK) // negedge SCK : SHIFT
	    begin
		if (r_count[2:0] == 7)  r_data_req   <= 1'b1;

		r_spi_out <= {r_spi_out[6:0], 1'b0 };
	    end
	else if (r_data_req & SPI_DATA_ACK) 
	    begin
		r_data_req   <= 1'b0;
		r_spi_out <= SPI_DATA_I;
	    end
end

endmodule
